# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for publishing or subscribing to a Kafka topic"""

import json
import logging
from datetime import datetime, timezone
from typing import Callable, Dict, List, Optional

import jsonschema
from kafka import KafkaConsumer, KafkaProducer
from pydantic import BaseSettings

from .utils import OutOfContextError


class EventSchemaNotFoundError(RuntimeError):
    """Thrown when an event schema was not found."""

    pass  # pylint: disable=unnecessary-pass


class EventSchemaValidationError(RuntimeError):
    """Thrown when event schema was not valid."""

    def __init__(self, reason: str):
        """Create error with corresponding message."""
        message = f"Event failed schema validation: {reason}"
        super().__init__(message)


# pylint: disable=too-few-public-methods
class KafkaConfigBase(BaseSettings):
    """A base class with config params related to
    asynchronous messaging.
    Inherit your config class from this class if you need
    to run an async PubSub API."""

    service_name: str
    client_suffix: str

    kafka_servers: List[str]


def validate_payload(
    payload: dict, json_schema: dict, raise_on_exception: bool = False
) -> bool:
    """Validate an event payload based on the specified json_schema."""
    try:
        jsonschema.validate(instance=payload, schema=json_schema)
        return True
    except jsonschema.exceptions.ValidationError as exc:
        logging.error(
            "%s: Message payload does not comform to JSON schema.",
            datetime.now(timezone.utc).isoformat(),
        )
        logging.exception(exc)
        if raise_on_exception:
            raise exc
        return False


def process_func_factory(
    exec_funcs: Dict[str, Callable[[dict], None]], event_schemas: Dict[str, dict]
):
    """
    Returns a function for processing all events that are of interest.
    It validates the payload and selects the correct function for processing out of
    the exec_funcs dict.
    """

    def process_func_wrapper(event):
        """Adds validation to exec_on_event function"""

        if "type" not in event.value.keys() or "payload" not in event.value.keys():
            raise EventSchemaValidationError(
                reason='The event value must contain the fields "type" and "payload".'
            )

        event_type = event.value["type"]

        if event_type in exec_funcs.keys():  # else ignore event
            json_schema = event_schemas[event_type]
            event_payload = event.value["payload"]
            validate_payload(
                event_payload, json_schema=json_schema, raise_on_exception=True
            )

            exec_func = exec_funcs[event_type]
            exec_func(event_payload)

    return process_func_wrapper


# pylint: disable=too-few-public-methods
class EventClient:
    """A base class to connect and iteract to/with a Kafka host."""

    def __init__(
        self,
        config: KafkaConfigBase,
        event_schemas: Dict[str, dict],
    ):
        """Initialize the Producer.

        Args:
            config [KafkaConfigBase]:
                Config paramaters provided as KafkaConfigBase object.
            event_schemas (Dict[str, dict]):
                A dictionary where the keys are the event types (== event keys) and the
                values are schemas for the event payload (== event value).
        """
        self.bootstrap_servers = config.kafka_servers
        self.service_name = config.service_name
        self.client_suffix = config.client_suffix
        self.client_id = f"{self.service_name}.{self.client_suffix}"
        self.event_schemas = event_schemas


class EventConsumer(EventClient):
    "A client to consume events from Apache Kafka."

    def __init__(
        self,
        config: KafkaConfigBase,
        topic_names: List[str],
        exec_funcs: Dict[str, Callable[[dict], None]],
        event_schemas: Dict[str, dict],
    ):
        """
        Initialize the EventConsumer.

        Args:
            config [KafkaConfigBase]:
                Config paramaters provided as KafkaConfigBase object.
            topic_names (List[str]):
                A list of topic name to subscribe to (only use letters, numbers, and "_").
            exec_funcs (Dict[str, Callable[[dict], None]]):
                A dictionary where the keys are the event types (== event keys) and the
                values are the callables to process the corresponding events.
                The only argument of the callables is the event value (a dict).
            event_schemas (Dict[str, dict]):
                A dictionary where the keys are the event types (== event keys) and the
                values are schemas for the event payload (== event value).
        """
        super().__init__(config=config, event_schemas=event_schemas)
        self.topic_names = topic_names
        self.exec_funcs = exec_funcs

        for event_type in exec_funcs.keys():
            if event_type not in event_schemas.keys():
                raise EventSchemaNotFoundError(
                    "The event_schemas dictionary is missing the schema for the event"
                    + f' type "{event_type}"'
                )

        self._consumer: Optional[KafkaConsumer] = None

    def __enter__(self):
        """Initialize the Kafka Producer"""

        self._consumer = KafkaConsumer(
            *self.topic_names,
            client_id=self.client_id,
            group_id=self.service_name,
            bootstrap_servers=self.bootstrap_servers,
            auto_offset_reset="earliest",
            key_deserializer=lambda event_key: event_key.decode("ascii"),
            value_deserializer=lambda event_value: json.loads(
                event_value.decode("ascii")
            ),
        )

        return self

    def __exit__(self, err_type, err_value, err_traceback):
        """
        Autocommit the current offset.
        """
        if not isinstance(self._consumer, KafkaConsumer):
            raise OutOfContextError()
        self._consumer.close(autocommit=True)

    def subscribe(self, run_forever: bool = True):
        """Subscribe to a topic and execute the specified function whenever
        a message is received.

        Args:
            run_forever (bool):
                If `True`, the function will continue to consume event for ever.
                If `False`, the function will wait for the first event, cosume it,
                and exit. Defaults to `True`.
        """

        if not isinstance(self._consumer, KafkaConsumer):
            raise OutOfContextError()

        process_event = process_func_factory(
            self.exec_funcs, event_schemas=self.event_schemas
        )

        if run_forever:
            for event in self._consumer:
                process_event(event)
        else:
            event = next(self._consumer)
            process_event(event)


class EventProducer(EventClient):
    "A client to publish events to Apache Kafka."

    def __init__(
        self,
        config: KafkaConfigBase,
        topic_name: str,
        event_schemas: Dict[str, dict],
    ):
        """Initialize the Producer.

        Args:
            config [KafkaConfigBase]:
                Config paramaters provided as KafkaConfigBase object.
            topic_name (str):
                The name of the topic (only use letters, numbers, and "_").
            event_schemas (Dict[str, dict]):
                A dictionary where the keys are the event types (== event keys) and the
                values are schemas for the event payload (== event value).
        """
        super().__init__(config=config, event_schemas=event_schemas)
        self.bootstrap_servers = config.kafka_servers
        self.service_name = config.service_name
        self.client_suffix = config.client_suffix
        self.client_id = f"{self.service_name}.{self.client_suffix}"
        self.topic_name = topic_name
        self.event_schemas = event_schemas

        self._producer: Optional[KafkaProducer] = None

    def __enter__(self):
        """Initialize the Kafka Producer"""

        self._producer = KafkaProducer(
            bootstrap_servers=self.bootstrap_servers,
            client_id=self.client_id,
            key_serializer=lambda event_key: event_key.encode("ascii"),
            value_serializer=lambda event_value: json.dumps(event_value).encode(
                "ascii"
            ),
        )

        return self

    def __exit__(self, err_type, err_value, err_traceback):
        """
        Flush and close producer. Blocks until all published events are transmitted.
        """
        if not isinstance(self._producer, KafkaProducer):
            raise OutOfContextError()
        self._producer.flush()
        self._producer.close()

    def publish(self, event_type: str, event_key: str, event_payload: dict):
        """Publish a message to the topic

        Args:
            event_type (str):
                The type of event. A keyword string that describes the action. Event
                types that reflect the lifetime of a user account could be:
                "user_registered", "user_activated", "user_inactived", "user_deleted",
                etc.
            event_key (str):
                The event key as str. Usually the identified of the resource that this
                event is related to. E.g. in a "user_registered" event, the key could be
                the ID of the user (e.g. "user-012323456").
            event_payload (dict):
                The event/message payload as a dict. I.e. the information that should be
                transmitted with this event.
        """

        if not isinstance(self._producer, KafkaProducer):
            raise OutOfContextError()

        # validate payload against schema:
        if event_type not in self.event_schemas.keys():
            raise EventSchemaNotFoundError(
                f'Schema for event of type "{event_type}"'
                + " was not found in the specified event_schemas dict."
            )
        event_schema = self.event_schemas[event_type]
        validate_payload(
            event_payload, json_schema=event_schema, raise_on_exception=True
        )

        # construct the event value:
        event_value = {
            "type": event_type,
            "payload": event_payload,
        }

        self._producer.send(self.topic_name, key=event_key, value=event_value)



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Protocols related to event handling."""

from typing import Protocol

from hexkit.custom_types import JSON


class EventPublisherProto(Protocol):
    """A protocol for publishing events to an event broker."""

    def publish(
        self, *, event_payload: JSON, event_type: str, event_key: str, topic_name: str
    ) -> None:
        """Publish an event."""
        ...



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the `akafka` module."""

from datetime import datetime

import pytest
from testcontainers.kafka import KafkaContainer

from hexkit.akafka import EventConsumer, EventProducer, KafkaConfigBase
from hexkit.utils import exec_with_timeout


class EventSuccessfullyConsumed(RuntimeError):
    """Raised when expected payload is received."""

    pass


def exec_dummy_func(received_payload: dict, expected_payload: dict):
    """Raises "EventSuccessfullyConsumed" exception if expected payload is received."""
    assert received_payload == expected_payload
    raise EventSuccessfullyConsumed()


def test_pub_sub():
    """Testing the Publisher and Subscriber classes."""
    topic_name = "test_topic"
    event_type = "test_event"
    event_key = "test_timestamp"
    event_payload = {"timestamp": datetime.now().isoformat()}
    event_schemas = {
        event_type: {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "additionalProperties": False,
            "properties": {
                "timestamp": {
                    "description": "The time when the message was published.",
                    "format": "date-time",
                    "type": "string",
                }
            },
            "required": ["timestamp"],
            "title": "test_schema",
            "type": "object",
        }
    }
    exec_funcs = {
        event_type: lambda received_payload: exec_dummy_func(
            received_payload, expected_payload=event_payload
        )
    }

    with KafkaContainer() as kafka:
        sub_config = KafkaConfigBase(
            service_name="test_sub",
            client_suffix="1",
            kafka_servers=[kafka.get_bootstrap_server()],
        )
        pub_config = KafkaConfigBase(
            service_name="test_pub",
            client_suffix="1",
            kafka_servers=[kafka.get_bootstrap_server()],
        )

        with EventProducer(
            config=pub_config, topic_name=topic_name, event_schemas=event_schemas
        ) as producer:
            producer.publish(
                event_type=event_type, event_key=event_key, event_payload=event_payload
            )

        with EventConsumer(
            config=sub_config,
            topic_names=[topic_name],
            exec_funcs=exec_funcs,
            event_schemas=event_schemas,
        ) as consumer:
            with pytest.raises(EventSuccessfullyConsumed):
                exec_with_timeout(
                    lambda: consumer.subscribe(run_forever=False), timeout_after=20000
                )



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Config fixtures"""

import copy
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from pydantic import BaseModel, BaseSettings

from hexkit.config import DEFAULT_CONFIG_PREFIX

from . import BASE_DIR, utils

# read all config yamls:
CONFIG_YAML_PATTERN = r"(.*)\.yaml"
CONFIG_YAML_DIR = BASE_DIR / "config_yamls"


class ConfigYamlFixture(BaseModel):
    """Container for config yaml fixtures"""

    path: Path
    content: Dict[str, Any]


def read_config_yaml(name: str):
    """Read config yaml."""
    path = CONFIG_YAML_DIR / name
    content = utils.read_yaml(path)

    return ConfigYamlFixture(path=path, content=content)


config_yamls = {
    re.match(CONFIG_YAML_PATTERN, cfile).group(1): read_config_yaml(cfile)  # type: ignore
    for cfile in os.listdir(CONFIG_YAML_DIR)
    if re.match(CONFIG_YAML_PATTERN, cfile)
}


# read env variable sets:
@dataclass
class EnvVarFixture:
    """Container for env var set. This class can be used
    as context manager so that the env vars are available
    within a with block but, after leaving the with block,
    the original enviroment is restored."""

    env_vars: Dict[str, str]
    prefix: str = DEFAULT_CONFIG_PREFIX

    def __enter__(self):
        """makes a backup of the environment and set the
        env_vars"""
        # pylint: disable=attribute-defined-outside-init
        self.env_backup = copy.deepcopy(os.environ)

        for name, value in self.env_vars.items():
            os.environ[f"{self.prefix}_{name}"] = value

    def __exit__(self, exc_type, exc_val, exc_tb):
        """restores the original environment"""
        os.environ = self.env_backup


def read_env_var_sets() -> Dict[str, EnvVarFixture]:
    """Read env vars sets and return a list of EnvVarFixtures."""
    env_var_dict = utils.read_yaml(BASE_DIR / "config_env_var_sets.yaml")

    return {
        set_name: EnvVarFixture(env_vars=env_vars)
        for set_name, env_vars in env_var_dict.items()
    }


env_var_sets = read_env_var_sets()


# pydantic BaseSettings classes:
class BasicConfig(BaseSettings):
    """Basic Config Example"""

    some_number: int
    some_string: str
    some_boolean: bool
    some_string_with_default: str = "default string"
    another_string_with_default: str = "another default string"



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General utilities that don't require heavy dependencies."""

import os
import signal
from pathlib import Path
from typing import Callable, Optional

from pydantic import BaseSettings

TEST_FILE_DIR = Path(__file__).parent.resolve() / "test_files"

TEST_FILE_PATHS = [
    TEST_FILE_DIR / filename
    for filename in os.listdir(TEST_FILE_DIR)
    if filename.startswith("test_") and filename.endswith(".yaml")
]


class OutOfContextError(RuntimeError):
    """Thrown when a context manager is used out of context."""

    def __init__(self):
        message = "Used context manager outside of a with block."
        super().__init__(message)


class DaoGenericBase:
    """A generic base for implementing DAO interfaces."""

    def __init__(self, config: BaseSettings):  # pylint: disable=unused-argument
        """Initialize DAO with a config params passed as pydantic BaseSettings object."""
        ...

    # Every DAO must support the context manager protocol.
    # However, it is up to the DAO implementation whether
    # the `__enter__` and `__exit__` functions are doing
    # something useful. They may remain stubs:

    def __enter__(self):
        """Setup logic."""
        ...
        return self

    def __exit__(self, err_type, err_value, err_traceback):
        """Teardown logic."""
        ...


# pylint: disable=too-few-public-methods
class AsyncDaoGenericBase:
    """A generic base for implementing an asynchronous DAO interfaces."""

    def __init__(self, config: BaseSettings):  # pylint: disable=unused-argument
        """Initialize DAO with a config params passed as pydantic BaseSettings object."""
        ...

    # Every DAO must support the context manager protocol.
    # However, it is up to the DAO implementation whether
    # the `__enter__` and `__exit__` functions are doing
    # something useful. They may remain stubs:

    async def __aenter__(self):
        """Setup logic."""
        ...
        return self

    # pylint: disable=unused-argument,no-self-use
    async def _aexit__(self, err_type, err_value, err_traceback):
        """Teardown logic."""
        ...


def raise_timeout_error(_, __):
    """Raise a TimeoutError"""
    raise TimeoutError()


def exec_with_timeout(
    func: Callable,
    timeout_after: int,
    func_args: Optional[list] = None,
    func_kwargs: Optional[dict] = None,
):
    """
    Exec a function (`func`) with a specified timeout (`timeout_after` in seconds).
    If the function doesn't finish before the timeout, a TimeoutError is thrown.
    """

    func_args_ = [] if func_args is None else func_args
    func_kwargs_ = {} if func_kwargs is None else func_kwargs

    # set a timer that raises an exception if timed out
    signal.signal(signal.SIGALRM, raise_timeout_error)
    signal.alarm(timeout_after)

    # execute the function
    result = func(*func_args_, **func_kwargs_)

    # disable the timer
    signal.alarm(0)

    return result


def create_fake_drs_uri(object_id: str):
    """Create a fake DRS URI based on an object id."""
    return f"drs://www.example.org/{object_id}"



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Config parsing functionality based on pydantic's BaseSettings"""

import os
from pathlib import Path
from typing import Any, Callable, Dict, Final, Optional

import yaml
from pydantic import BaseSettings

# Default config prefix:
DEFAULT_CONFIG_PREFIX: Final = "ghga_services"


class ConfigYamlDoesNotExist(RuntimeError):
    """Thrown when the context manager is used out of context."""

    def __init__(self, path: Path, specified_via: Optional[str] = None):
        message = (
            "The config yaml " + ""
            if specified_via is None
            else f"specified via {specified_via} " + f"does not exist: {path}"
        )
        super().__init__(message)


def get_default_config_yaml(prefix: str) -> Optional[Path]:
    """Get the path to the default config function.

    Args:
        prefix (str):
            Name prefix used to derive the default paths.
    """
    # check if path to config is set as env variable:
    path_env_var = f"{prefix.upper()}_CONFIG_YAML"
    try:
        path_via_env_var = Path(os.environ[path_env_var])
        if path_via_env_var.is_file():
            return path_via_env_var
        raise ConfigYamlDoesNotExist(
            path=path_via_env_var,
            specified_via=f"the environment variable {path_env_var}",
        )
    except KeyError:
        pass

    # construct file name from prefix:
    file_name = f".{prefix}.yaml"

    # look in the current directory:
    path_in_cwd = Path(os.getcwd()) / file_name
    if os.path.isfile(path_in_cwd):
        return path_in_cwd

    # look in the home directory:
    path_in_home = Path.home() / file_name
    if os.path.isfile(path_in_home):
        return path_in_home

    # if nothing was found return None:
    return None


def yaml_settings_factory(
    config_yaml: Optional[Path] = None,
) -> Callable[[BaseSettings], Dict[str, Any]]:
    """
    A factory of source methods for pydantic's BaseSettings Config that load
    settings from a yaml file.

    Args:
        config_yaml (str, Optional):
            Path to the yaml file to read from.
    """

    def yaml_settings(  # pylint: disable=unused-argument
        settings: BaseSettings,
    ) -> Dict[str, Any]:
        """source method for loading pydantic BaseSettings from a yaml file"""
        if config_yaml is None:
            return {}

        with open(config_yaml, "r", encoding="utf8") as yaml_file:
            return yaml.safe_load(yaml_file)

    return yaml_settings


def config_from_yaml(
    prefix: str = DEFAULT_CONFIG_PREFIX,
) -> Callable:
    """A factory that returns decorator functions which extends a
    pydantic BaseSettings class to read in parameters from a config yaml.
    It replaces (or adds) a Config subclass to the BaseSettings class that configures
    the priorities for parameter sources as follows (highest Priority first):
        - parameters passed using **kwargs
        - environment variables
        - file secrets
        - yaml config - path specified via env variable `{prefix}_CONFIG_YAML`
                        (all uppercase)
        - yaml config - file named `.{prefix}.yaml` and placed in the CWD or home dir
        - defaults

    Args:
        prefix: (str, optional):
            When defining parameters via enviroment variables, all variables
            have to be prefixed with this string following this pattern
            "{prefix}_{actual_variable_name}". Moreover, this prefix is used
            to derive the default location for the config yaml file
            ("~/.{prefix}.yaml"). Defaults to "ghga_services".
    """

    def decorator(settings) -> Callable:
        # settings should be a pydantic BaseSetting,
        # there is no type hint here to not restrict
        # autocompletion for attributes of the
        # modified settings class returned by the
        # contructor_wrapper
        """The actual decorator function.

        Args
            settings (BaseSettings):
                A pydantic BaseSettings class to be modified.
        """

        # check if settings inherits from pydantic BaseSettings:
        if not issubclass(settings, BaseSettings):
            raise TypeError(
                "The specified settings class is not a subclass of pydantic.BaseSettings"
            )

        def constructor_wrapper(
            config_yaml: Optional[Path] = None,
            **kwargs,
        ):
            """A wrapper for constructing a pydantic BaseSetting with modified sources

            Args:
                config_yaml (str, optional):
                    Path to a config yaml. Overwrites the default location.
            """

            # get default path if config_yaml not specified:
            if config_yaml is None:
                config_yaml = get_default_config_yaml(prefix)
            else:
                if not config_yaml.is_file():
                    raise ConfigYamlDoesNotExist(path=config_yaml)

            class ModSettings(settings):  # type: ignore
                """Modifies the orginal Settings class provided by the user"""

                # pylint: disable=too-few-public-methods
                class Config:
                    """pydantic Config subclass"""

                    # add this prefix to all variable names to
                    # define them as environment variables:
                    env_prefix = f"{prefix}_"

                    @classmethod
                    def customise_sources(
                        cls,
                        init_settings,
                        env_settings,
                        file_secret_settings,
                    ):
                        """add custom yaml source"""
                        return (
                            init_settings,
                            env_settings,
                            file_secret_settings,
                            yaml_settings_factory(config_yaml),
                        )

            # construct settings class:
            return ModSettings(**kwargs)

        return constructor_wrapper

    return decorator



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Apache Kafka specific event publishing provider.

Require dependencies of the `akafka` extra. See the `setup.cfg`.
"""

import json

from kafka import KafkaProducer

from hexkit.custom_types import JSON
from hexkit.eventpub.protocol import EventPublisherProto


class KafkaEventPublisher(EventPublisherProto):
    """Apache Kafka specific event publishing provider."""

    def __init__(self, service_name: str, client_suffix: str, kafka_servers: list[str]):
        """Initialize the provider with some config params.

        Args:
            service_name (str):
                The name of the (micro-)service from which messages are published.
            client_suffix (str):
                String that uniquely this instance across all instances of this service.
                Will create a globally unique Kafka client IDidentifier by concatenating
                the service_name and the client_suffix.
            kafka_servers (list[str]):
                List of connection strings pointing to the kafka brokers.
        """

        self.client_id = f"{service_name}.{client_suffix}"

        self._producer = KafkaProducer(
            bootstrap_servers=kafka_servers,
            client_id=self.client_id,
            key_serializer=lambda event_key: event_key.encode("ascii"),
            value_serializer=lambda event_value: json.dumps(event_value).encode(
                "ascii"
            ),
        )

    def publish(
        self, *, event_payload: JSON, event_type: str, event_key: str, topic_name: str
    ) -> None:
        """Publish event to Kafka broker."""

        event_value = {
            "type": event_type,
            "payload": event_payload,
        }
        self._producer.send(topic_name, key=event_key, value=event_value)
        self._producer.flush()



# Copyright 2021 - 2022 Universität Tübingen, DKFZ and EMBL
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test config parsing module"""
import os
import shutil
from pathlib import Path

import pytest

from hexkit.config import config_from_yaml

from .fixtures.config import BasicConfig, config_yamls, env_var_sets


def test_config_from_yaml():
    """Test that config yaml correctly overwrites
    default parameters"""

    config_yaml = config_yamls["basic"]

    # update config class with content of config yaml
    config_constructor = config_from_yaml()(BasicConfig)
    config = config_constructor(config_yaml=config_yaml.path)

    # compare to expected content:
    expected = BasicConfig(**config_yaml.content)
    assert config.dict() == expected


def test_config_from_env():
    """Test that env vars correctly overwrites
    default parameters"""
    env_var_fixture = env_var_sets["basic_complete"]
    with env_var_fixture:
        # update config class with content of config yaml and
        # from the env vars
        config_constructor = config_from_yaml()(BasicConfig)
        config = config_constructor()

    # compare to expected content:
    expected = BasicConfig(**env_var_fixture.env_vars)
    assert config.dict() == expected


def test_config_from_yaml_and_env():
    """Test that config yaml and env vars correctly overwrites
    default parameters"""

    config_yaml = config_yamls["basic"]
    env_var_fixture = env_var_sets["basic_partly"]

    with env_var_fixture:
        # update config class with content of config yaml and
        # from the env vars
        config_constructor = config_from_yaml()(BasicConfig)
        config = config_constructor(config_yaml=config_yaml.path)

    # compare to expected content:
    overwrite_params = {**config_yaml.content, **env_var_fixture.env_vars}
    expected = BasicConfig(**overwrite_params)
    assert config.dict() == expected


@pytest.mark.parametrize("cwd", [True, False])
def test_config_from_default_yaml(cwd: bool):
    """Test that default config yaml from home is correctly read"""

    base_dir = Path(os.getcwd()) if cwd else Path.home()
    prefix = "test_prefix"

    # copy basic config to default config location:
    config_yaml = config_yamls["basic"]
    default_yaml_path = base_dir / f".{prefix}.yaml"
    shutil.copy(config_yaml.path, default_yaml_path)

    # update config class with content of config yaml
    config_constructor = config_from_yaml(prefix=prefix)(BasicConfig)
    config = config_constructor()

    # cleanup default config yaml:
    os.remove(default_yaml_path)

    # compare to expected content:
    expected = BasicConfig(**config_yaml.content)
    assert config.dict() == expected


def test_config_from_default_yaml_via_env():
    """Test that default config yaml specified via an environment variable is correctly
    read"""

    prefix = "test_prefix"

    # set env var:
    config_yaml = config_yamls["basic"]
    os.environ[f"{prefix.upper()}_CONFIG_YAML"] = str(config_yaml.path)

    # update config class with content of config yaml
    config_constructor = config_from_yaml(prefix=prefix)(BasicConfig)
    config = config_constructor()

    # compare to expected content:
    expected = BasicConfig(**config_yaml.content)
    assert config.dict() == expected


def test_error_on_invalid_base_class():
    """Check that an exception is thrown if the class passed to the
    config decorator does not inherit from pydantic.BaseSettings
    """

    class WrongClass:
        """Not inheriting from pydantic.BaseSettings"""

        pass  # pylint: disable=unnecessary-pass

    with pytest.raises(TypeError):
        config_from_yaml()(WrongClass)()
